import java.awt.*;
import java.awt.image.*;
import java.util.*;

/**
 * pointer - Text and sliding Triangles with automatic timedependant slowdown and fading<br>
 * 
 * @author 	<A HREF="http://www.sbk-ks.de/~twe/e">Thomas Wendt</A>
 * @version 	public
 */
public class pointer extends java.applet.Applet implements Runnable {
  /** The message to be displayed. */
  String message;
 
  /** x-position of message. */
  int messageX;

  /** y-position of message. */
  int messageY;

  /** Image to create with the triangles. */
  Image img;

  /** x-position of image. */
  int imgX;

  /** Minimal x-position. */
  int imgXmin;

  /** The animating thread. */
  Thread t;

  /** Animation speed: position increment. */
  int iSpeed;

  /** Animation speed: Threads pause. */
  int tSpeed;

  /** The direction of animation. */
  int dir;

  /** The Text-color. */
  Color txtCo;

  /** The color of the triangles. */
  Color fgCo;

  /** The backgroundcolor. */
  Color bgCo;

  /** The color to use after expiring. */
  Color expCo;

  /**
   * The age of the applet
   * 1 = new, 0 = old
   */
  float age = 1.0f;

  /**
   * The time-dependant state of the applet
   * True, if expired
   */
  boolean expired;

  /**
   * The size used to calc the Font.
   * Makes resizing possible.
   */
  Dimension lastS = new Dimension(1,1);

  /**
   * Initialize: Read Attributes
   * Calc speeds to give other threads a chance.
   * Resizes to (2,2) and do nothing if expired;
   */
  final public void init () {
    String at = getParameter("msg");
    message = (at == null) ? "Neu!" : at;

    // silly, but Color matches Date very well
    fgCo = readColor(getParameter("exp"), Color.white);
    if (!fgCo.equals(Color.white)) {
      Date exp = new Date(fgCo.getRed(),fgCo.getGreen()-1,fgCo.getBlue());
      Date today = new Date();
      expired = today.after(exp);
      fgCo = readColor(getParameter("creat"), Color.white);
      if (!fgCo.equals(Color.white)) { 
        Date creat = new Date(fgCo.getRed(),fgCo.getGreen()-1,fgCo.getBlue());
        float diff = exp.getTime() - creat.getTime();
        age = exp.getTime() - today.getTime();
        age = age/diff;
        if (age<0.01f)
          age = 0.001f;
      }
    }

    // dont show, if expired
    expCo = readColor(getParameter("expco"), getBackground());
    if (expired) {
      resize(2,2);
      return;
    }

    // calc speeds to let others a chance
    at = getParameter("speed");
    tSpeed = (at == null) ? 100 : (Integer.valueOf(at).intValue());
    dir = (tSpeed > 0) ? 1 : -1;
    tSpeed = new Float(1000*dir/(tSpeed*age)).intValue();
    if (tSpeed < 1) { tSpeed = 1; }
    iSpeed = dir*(1+20/tSpeed);
    tSpeed = dir*iSpeed*tSpeed;
    
    txtCo = readColorAge(getParameter("txtco"), Color.blue);
    fgCo = readColorAge(getParameter("fgco"), getBackground());
    bgCo = readColorAge(getParameter("bgco"), Color.yellow);

  }

  /** Parameter Info. */
  final public String[][] getParameterInfo() {
    String[][] info = {
      {"msg",    "String",  "Message to display (New!)"},
      {"speed",  "int",     "animation speed (1/msec)"},
      {"txtco",  "int[3]",  "RGB-Color of Message (blue)"},
      {"fgco",   "int[3]",  "RGB-Color of triangles (getBackground)"},
      {"bgco",   "int[3]",  "RGB-Color of background (yellow)"},
      {"expco",  "int[3]",  "RGB-Color to use if expired (getBackground)"},
      {"creat",  "int[3]",  "Date of creation: Y, M, D; if not set, no slowdown, fading"},
      {"exp",    "int[3]",  "Date to expire: Y, M, D; if not set, no slowdown, fading, expiration"},
    };
    return info;
  }

  /** Applet Info. */
  final public String getAppletInfo() {
    return "pointer.java by Thomas Wendt, http://www.sbk-ks.de/~twe/e";
  }    

  /**
   * Convert a ','-delimited String with RGB-Values to Color
   * Uses aDefault, if no or not enough RGB-Values
   */
  final Color readColor(String aColor, Color aDefault) {
    if (aColor == null) { return aDefault; }

    int r, g, b;
    StringTokenizer st = new StringTokenizer(aColor, ",");
    
    try {
      r = Integer.valueOf(st.nextToken()).intValue();
      g = Integer.valueOf(st.nextToken()).intValue();
      b = Integer.valueOf(st.nextToken()).intValue();
      return new Color(r,g,b);
    }
    catch (Exception e) { return aDefault; }
  }

  /** Compute a given color component with age. */
  final int coAge(float fg, float bg) {
    return new Float(fg*age+(1-age)*bg).intValue();
  }

  /**
   * Convert by calling readColor and take age into account
   * @see #readColor
   * @see #coAge
   */
  final Color readColorAge(String aColor, Color aDefault) {
    Color co = readColor(aColor, aDefault);
    int r = Color.lightGray.getRed();
    int g = Color.lightGray.getGreen();
    int b = Color.lightGray.getBlue();
    
    return new Color(coAge(co.getRed(),r), coAge(co.getGreen(), g), coAge(co.getBlue(), b));
  }

  /** Create the image offline. */
  final void createImg() {
    // set some constants
    int w = size().width;
    int h = size().height;
    lastS.width = w;
    lastS.height = h;

    // calc font: message must fit in applets area
    int refH = 14;
    Font tf = new Font("TimesRoman", Font.BOLD, refH);
    setFont(tf);
    FontMetrics tfm = getFontMetrics(tf);
    int fw = tfm.stringWidth(message);
    int fh = tfm.getHeight();
    fh = Math.min(refH*h/fh, refH*w/(fw << 1));
    Font f = new Font("TimesRoman", Font.BOLD, fh);
    FontMetrics fm = getFontMetrics(f);
    fw = fm.stringWidth(message);
    fh = fm.getHeight();
    setFont(f);
    messageX = (w-fw) >> 1;
    messageY = ((h-fh) >> 1)+fm.getAscent();

    // then draw offline image
    w = ((w/h)+2)*h; // round up to multipl of height
    int pixels[] = new int[w*h];
    int x;
    int y;

    // fill with background
    int co = (255 << 24) | (bgCo.getRed() << 16) | (bgCo.getGreen() << 8) | bgCo.getBlue();
    for (x = 0; x < (w*h); x++) { pixels[x] = co; }

    co = (255 << 24) | (fgCo.getRed() << 16) | (fgCo.getGreen() << 8) | fgCo.getBlue();
      
    /* 
     * Create triangles
     * Using bytearray might be faster than using offscreen Graphics
     */
    for (x = h >> 1; x < w; x += h) {
      for (y = 1; y <= (h >> 1); y++) {
        for (int i = x; (i <= (x+y)) && (i >= (x-y)) ; i += dir) {
          pixels[y*w+i] = co;
          pixels[(h-y-1)*w+i] = co;
        }
      }
    }
    // create an ImageProducer with the bytearray
    img = createImage(new MemoryImageSource(w, h, pixels, 0, w));
    imgXmin = -h;
    imgX = (dir > 0) ? imgXmin : 0;
  }

  /** Show the stuff, call paint */
  final public void update(Graphics g) { paint(g); }

  /** Show the stuff */
  final public void paint(Graphics g) {
    // do nothing, if expired
    if (expired) {
      if (expCo != null) {
        g.setColor(expCo);
        g.fillRect(0,0,size().width,size().height);
      }
      return;
    }

    // if something has changed, create new image
    if ((img == null) || (size().height != lastS.height) || (size().width != lastS.width)) { createImg(); }

    // if position in bounds
    if ((imgX <= 0) && (imgX >= imgXmin)) {
      // draw first moving image
      g.drawImage(img,imgX,0, this);
      g.setColor(txtCo);
      // then text
      g.drawString(message, messageX, messageY);
    }
  }

  /** Run the loop. This method is called by class Thread. */
  final public void run() {
    // do nothing, if expired
    if (expired)
     return;

    // others may be more important
    t.setPriority(t.MIN_PRIORITY);
    while (t != null) {
      // increment position
      imgX += iSpeed;
      if (imgX > 0)
        imgX = imgXmin;
      else if (imgX < imgXmin)
        imgX = 0;
      // paint the applet
      repaint();
      // and pause
      try { Thread.sleep(tSpeed); }
      catch (Exception e) { }
    }
  }

  /** Start the applet by forking an animation thread. */
  final public void start() {
    if (t == null) {
      t = new Thread(this);
      t.start();
    }
  }

  /** Stop the applet. The thread will exit because run() exits. */
  final public void stop() {
    if (t != null)
      t.stop();
    t = null;
  }
}
